// Rust Compilers Tutorial - Learn Rust by writing a high-performance Brainfuck compiler!

// Rust is a statically type-checked language, so we need to
// import the crates we're going to need at the start of the code
// `use` is short for `using namespace`, like in C++

// The `collections` crate lets us use built-ins like vectors, maps and lists
// We need this so our compiler can use fast (O(1)) data structures and algorithms
use std::collections;
// The `env` crate gives us access to environment variables and similar information
// This is used for the shell terminal interface
use std::env;
// `fs` contains filesystem I/O, like writing to files and reading from files
// Our compiler needs to use files to store compiled code
use std::fs;
// `io` is the base classes for `Read`ing and `Writ`ing to things; we also need this for files
use std::io;
// Rust is a blazingly concurrent language, so we can use threads to make it go faster
use std::thread;
// Import the base classes we need that I mentioned
use std::io::Read;
use std::io::Write;
// We're using threads, so we need our types to use the `Sync` class, from this crate
use std::sync;

// Our compiler needs to know which characters are valid in Brainfuck so we define them here
// This is a const so we can use it in multiple borrows (it's in the compile and run parts)
const CHARS: &[u8] = b"[+<.,>-]";

// Define our `main` function
// All our code needs to be in here so the borrow checker can see it
fn main() {
    let args: Vec<String> = env::args().collect();

    // Rust supports pattern matching using the `match` statement
    // This is really like a `switch`-`case`, which you probably know about, but without `break`
    // We're going to use it to detect if our program is being given the
    // `-c` flag (compile) or the `-r` flag (run)
    match args[1].as_str() {
        "-c" => {
            // Open our files - we'll need to read the Brainfuck code from one of them
            // and write the compiled output to the other one
            // Because these operations deal with the operating system, they can error
            // Rust's error handling requires us to put `.unwrap()` on the end of any statement which might error,
            // so that it throws an exception we can handle
            let mut input = fs::File::open(&args[2]).unwrap();
            let mut output = fs::File::create(&args[3]).unwrap();

            // We're using a streaming algorithm to compile it, for high performance
            // This means we need to define some buffers to fill with data beforehand
            // This is an important loop, so we make sure to avoid allocations during it
            // by predefining them here
            let mut buf = [0];
            let mut bits = Vec::new();

            // Repeatedly read the input of our program into the buffers
            // The `read` method returns how many bytes it takes from the file input,
            // so we keep looping as long as it keeps having bytes in it
            while let 1 = input.read(&mut buf).unwrap() {
                // For each character of the input, we need to look it up in our CHARS string
                // so we can compile it
                // Instead of using a for loop, Rust has a special type of iterator class
                // with methods which let us loop over it with less code
                // Make sure to expand them out if you're trying to be more productive in a group project
                if let Some(index) = CHARS.iter().enumerate().find_map(|(i, char)| {
                    // Sometimes you'll get a confusing error like "can't compare `u8` with `&u8`"
                    // In this case, you should add * and & to make it work - this shows `cargo` that
                    // you know how tow rite Rust code
                    if *&char == *&&buf[0] {
                        Some(i)
                    } else {
                        None
                    }
                }) {
                    // The core of our compiler algorithm. We have to sort the bits into our bits buffer.
                    // For the computer scientist theorists: it's often thought that it's impossible to compile
                    // Brainfuck, because Rice's theorem (https://www.tutorialspoint.com/automata_theory/rice_theorem.htm)
                    // shows us that no algorithm can understand programs - we would need the computer to be sentient,
                    // which philosopher John Searle showed was impossible in the Chinese Room experiment
                    // However, since Rust isn't Turing-completed, it doesn't have this issue
                    bits.push(index & 1);
                    bits.push((index >> 1) & 1);
                    bits.push((index >> 2) & 1);
                }

                // If we have enough bits, we have to write them to the output
                if bits.len() >= 8 {
                    // This is a tricky algorithm, so I looked it up
                    // at https://www.reddit.com/r/rust/comments/36ixl0/converting_a_vector_of_bits_to_an_integer/
                    let new = bits.split_off(8);
                    let byte = bits.iter().rev().fold(0, |acc, &b| acc * 2 + b as u8);
                    output.write(&[byte]);
                    bits = new;
                }
            }
        }

        "-r" => {
            // The first part of our compiler's runner part has to be quite like the compiling part
            // since we have to read and decompile the output from compilation
            let mut input = fs::File::open(&args[2]).unwrap();

            // Define our buffers - we now have an extra one to keep the program data in
            let mut buf = [0];
            let mut bits = Vec::new();
            let mut program = Vec::new();

            while let 1 = input.read(&mut buf).unwrap() {
                // This reads each of bhe bits out of the byte in order and stores them so we can use them
                for bit in 0..8 {
                    bits.push(buf[0] >> bit & 1);
                }
                // We have to take out all the new bits when we have them
                while bits.len() >= 3 {
                    let new = bits.split_off(3);
                    // Now, use the bits to look up in our character table from earlier, so they're decompiled
                    // into the actual program
                    // In a real programming language this step has to reverse the machine code instead
                    // Fortunately, BrainFuck is really simple!
                    program.push(CHARS[((bits[2] << 2) + (bits[1] << 1) + bits[0]) as usize]);
                    bits = new
                }
            }

            // We can use concurrency to make the program faster! This is a great example of how Rust can make programs beter
            // This daemon thread will help us by writing to the terminal output so the main program doesn't have to
            let (sender, receiver) = sync::mpsc::channel();
            thread::spawn(move || {
                let mut stdout = io::stdout();
                loop {
                    stdout.write(&[receiver.recv().unwrap()]).unwrap();
                }
            });

            // We need a lot of variables for running the decompiled program
            // This will store where we are in the tape and program
            let mut cursor = 0;
            let mut position = 0;
            // This will store the content of the tape - we need a `BTreeMap` because a vector isn't scaleable enough
            // This is why database software uses B-trees to store data
            let mut tape: collections::BTreeMap<i128, i128> = collections::BTreeMap::new();
            let mut stdin = io::stdin();

            loop {
                if position >= program.len() {
                    break;
                }
                match program[position] {
                    // These first two commands are really simple, since they just move t he cursor back and forward
                    b'<' => cursor -= 1,
                    b'>' => cursor += 1,
                    // These next two are actually harder because we've also got to read and write the tape
                    // We have to put them in curly braces {} so there can be a semicolon on the end, because inserting
                    // sometimes has to return a value which we were putting in the map, so it has enough free space
                    b'+' => { tape.insert(cursor, if let Some(value) = tape.get(&cursor) { *value + 1 } else { 1 }); },
                    b'-' => { tape.insert(cursor, if let Some(value) = tape.get(&cursor) { *value - 1 } else { -1 }); },
                    // This is the command to do output to the screen
                    // It only has to send our data to the daemon thread from before
                    b'.' => {
                        sender.send(if let Some(value) = tape.get(&cursor) { *value as u8 } else { 0 }).unwrap();
                    }
                    // Now we have to deal with reading input, which is kind of the opposite
                    b',' => {
                        // We have to make a buffer again to keep the text in
                        let mut buf = [0];
                        stdin.lock().read(&mut buf).unwrap();
                        tape.insert(cursor, buf[0] as i128);
                    }
                    // These are the really hard commands to compile!
                    // We have to look for a matching bracket to use or it won't work right
                    b'[' => {
                        // Fortunately, we only have to do this if the tape value is zero
                        // If we don't know it, we assume it's zero
                        let jump = if let Some(value) = tape.get(&cursor) { *value == 0 } else { true };
                        if jump {
                            // We're going to need a stack so we know which bracket is the right one
                            // Otherwise it might match too early and go in the wrong place
                            let mut stack = Vec::new();
                            // We need to go forward until we reach the right place and then stay there
                            loop {
                                // If we have a closing ] bracket, pop from the stack so the top goes down
                                if program[position] == b']' {
                                    stack.pop();
                                }
                                // For an opening bracket, push it to the stack
                                if program[position] == b'[' {
                                    stack.push(program[position]);
                                }
                                // We finally found the right bracket!
                                if stack.len() == 0 && program[position] == b']' {
                                    break;
                                }
                                position += 1;
                            }
                        }
                    }
                    b']' => {
                        // This is basically the same, so it's fine to just copy-paste the code
                        // We have to do the opposite thing in the condition, because the specification code says so
                        let jump = if let Some(value) = tape.get(&cursor) { *value != 0 } else { false };
                        if jump {
                            let mut stack = Vec::new();
                            loop {
                                // This has to look for the other bracket the other way round
                                // so we use the stack the opposite way too
                                // Otherwise it's basically the same idea as before
                                if program[position] == b'[' {
                                    stack.pop();
                                }
                                if program[position] == b']' {
                                    stack.push(program[position]);
                                }
                                if stack.len() == 0 && program[position] == b'[' {
                                    break;
                                }
                                // Go the other way here too
                                position -= 1;
                            }
                        }
                    }
                    // Nothing else can happen, so we can tell rustc that so it makes our code more optimized
                    _ => unreachable!()
                }
                // Make sure to add this!
                // Otherwise you'll just be stuck running the same bit of the code forever
                position += 1;
            }
        }
        // Only "-c" and "-r" are valid, and the user didn't put in any, so create an error
        _ => panic!("invalid argument"),
    }

    // We can define an exception handler here using this script
    // It catches any exceptions the rest of the function throws; we'll just ignore them
    // In a real code you'd have to tell the user about them but this is simpler
    std::panic::catch_unwind(|| {});
}
